import 'package:flutter/painting.dart';

import '../extensions/extensions.dart';
import 'matrix_renderer.dart';

import 'anchor.dart';
import 'cache_renderer_mixin.dart';
import 'renderer.dart';

class Actor extends MatrixRenderer with CacheRendererMixin {
  Actor({
    double? angle,
    Vector2? position,
    Vector2? skew,
    Vector2? size,
    Vector2? scale,
    Anchor? anchor,
    double? priority,
  }) : super(
          angle: angle,
          position: position,
          skew: skew,
          size: size,
          anchor: anchor,
          scale: scale,
          priority: priority,
        );

  @override
  bool containsPoint(Vector2 point) {
    final local = absoluteToLocal(point);
    return (local.x >= 0) && (local.y >= 0) && (local.x < size.x) && (local.y < size.y);
  }

  Vector2 parentPositionOf(Vector2 point) {
    return MatrixUtils.transformPoint(matrix, point.toOffset()).toVector2();
  }

  Vector2 parentPositionOfAnchor(Anchor anchor) {
    return anchor == this.anchor
        ? position
        : parentPositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 absolutePositionOf(Vector2 point) {
    Vector2 parentPoint = parentPositionOf(point);
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        parentPoint = ancestor.parentPositionOf(parentPoint);
      }
      ancestor = ancestor.parent;
    }
    return parentPoint;
  }

  Vector2 absolutePositionOfAnchor(Anchor anchor) {
    return absolutePositionOf(Vector2(anchor.x * size.x, anchor.y * size.y));
  }

  Vector2 toLocal(Vector2 point) {
    final transform = matrix.clone();
    final det = transform.invert(); // side effect

    if (det == 0.0) {
      return Vector2.zero();
    }

    final Vector3 n = Vector3(0.0, 0.0, 1.0);
    final Vector3 i = transform.perspectiveTransform(Vector3(0.0, 0.0, 0.0));
    final Vector3 d = transform.perspectiveTransform(Vector3(0.0, 0.0, 1.0)) - i;
    final Vector3 s = transform.perspectiveTransform(Vector3(point.x, point.y, 0.0));
    final Vector3 p = s - d * (n.dot(s) / n.dot(d));

    return Vector2(p.x, p.y);
  }

  Vector2 absoluteToLocal(Vector2 point) {
    Renderer? ancestor = parent;

    while (ancestor != null) {
      if (ancestor is Actor) {
        return toLocal(ancestor.absoluteToLocal(point));
      }
      ancestor = ancestor.parent;
    }
    return toLocal(point);
  }

  Rect toRect() => _toRectImpl(parentPositionOfAnchor);

  Rect toAbsoluteRect() => _toRectImpl(absolutePositionOfAnchor);

  Rect _toRectImpl(Vector2 Function(Anchor point) projector) {
    final topLeft = projector(Anchor.topLeft);
    final bottomRight = projector(Anchor.bottomRight);

    if (angle == 0.0) {
      return Rect.fromPoints(topLeft.toOffset(), bottomRight.toOffset());
    }

    final topRight = projector(Anchor.topRight);
    final bottomLeft = projector(Anchor.bottomLeft);
    final xs = [topLeft.x, topRight.x, bottomLeft.x, bottomRight.x]..sort();
    final ys = [topLeft.y, topRight.y, bottomLeft.y, bottomRight.y]..sort();
    return Rect.fromLTRB(xs.first, ys.first, xs.last, ys.last);
  }

  double get absoluteAngle {
    return ancestors()
        .whereType<MatrixRenderer>()
        .fold<double>(angle, (totalAngle, c) => totalAngle + c.angle);
  }

  Vector2 get absolutePosition => absolutePositionOfAnchor(anchor);
  Vector2 get center => parentPositionOfAnchor(Anchor.center);
  Vector2 get absoluteCenter => absolutePositionOfAnchor(Anchor.center);
}
