import 'package:flutter/foundation.dart';
import 'package:flutter/painting.dart';

import '../extensions/extensions.dart';
import 'renderer.dart';
import 'anchor.dart';

abstract class MatrixRenderer extends Renderer {
  final Matrix4 _matrix = Matrix4.identity();
  bool _matrixCached = false;

  double _angle;
  Anchor _anchor;
  Vector2 _size;
  Vector2 _position;
  Vector2 _skew;
  Vector2 _scale;

  MatrixRenderer({
    double? angle,
    Vector2? position,
    Vector2? skew,
    Vector2? size,
    Anchor? anchor,
    Vector2? scale,
    double? priority,
  })  : _angle = angle ?? 0,
        _anchor = anchor ?? Anchor.center,
        _position = position ?? Vector2.zero(),
        _skew = skew ?? Vector2.zero(),
        _size = size ?? Vector2.zero(),
        _scale = scale ?? Vector2.all(1),
        super(priority: priority);

  double get angle => _angle;
  set angle(double next) {
    if (_angle != next) {
      onAngleChanged(_angle, _angle = next);
    }
  }

  @mustCallSuper
  void onAngleChanged(double prev, double next) {
    refresh();
    _matrixCached = false;
  }

  Vector2 get position => _position;
  set position(Vector2 next) {
    if (_position != next) {
      onPositionChanged(_position, _position = next.clone());
    }
  }

  @mustCallSuper
  void onPositionChanged(Vector2 prev, Vector2 next) {
    refresh();
    _matrixCached = false;
  }

  Vector2 get scale => _scale;
  set scale(Vector2 next) {
    if (_scale != next) {
      onScaleChanged(_scale, _scale = next.clone());
    }
  }

  @mustCallSuper
  void onScaleChanged(Vector2 prev, Vector2 next) {
    refresh();
    _matrixCached = false;
  }

  Vector2 get skew => _skew;
  set skew(Vector2 next) {
    if (_skew != next) {
      onSkewChanged(_skew, _skew = next.clone());
    }
  }

  @mustCallSuper
  void onSkewChanged(Vector2 prev, Vector2 next) {
    refresh();
    _matrixCached = false;
  }

  Vector2 get size => _size;
  set size(Vector2 next) {
    if (_position != next) {
      onSizeChanged(_size, _size = next.clone());
      refresh();
    }
  }

  @mustCallSuper
  void onSizeChanged(Vector2 prev, Vector2 next) {
    refresh();
    _matrixCached = false;
  }

  Anchor get anchor => _anchor;
  set anchor(Anchor next) {
    if (_anchor != next) {
      onAnchorChanged(_anchor, _anchor = next);
      refresh();
    }
  }

  @mustCallSuper
  void onAnchorChanged(Anchor prev, Anchor next) {
    refresh();
    _matrixCached = false;
  }

  Matrix4 get matrix {
    if (_matrixCached) {
      return _matrix;
    }

    final offsetX = _anchor.x * _size.x;
    final offsetY = _anchor.y * _size.y;

    _matrix
      ..setIdentity()
      ..translate(_position.x, _position.y);

    if (_angle != 0) {
      _matrix.rotateZ(_angle);
    }

    if (_scale.x != 1 || _scale.y != 1) {
      _matrix.scale(_scale.x, _scale.y);
    }

    if (offsetX != 0 || offsetY != 0) {
      _matrix.translate(-offsetX, -offsetY);
    }

    if (_skew.x != 0 || _skew.y != 0) {
      _matrix.multiply(Matrix4.skew(_skew.x, _skew.y));
    }

    _matrixCached = true;
    return _matrix;
  }

  @override
  void transformCanvas(Canvas canvas, VoidCallback callback) {
    canvas
      ..save()
      ..transform(matrix.storage);
    callback();
    canvas.restore();
  }
}
